// Control para el motor NIDEC 24H677
#include <math.h>

// Definición de constantes
#define Ts 10 // Tiempo de muestreo  para el Arduino
#define Tsc 0.01 // Tiempo de muestreo para el PID
#define Kp 90 // Constante proporcional
#define Ki 35 // Constante integral
#define Kc 0.00001 // Constante de corrección de velocidad en posición 0
#define VMAX 2400 // Máxima salida pwm permisible
#define DATMIN -5.0
#define DATMAX 5.0
#define FREC 0.05

// Definiciones de los pines
#define Pin_pend A1 // POSICION
#define Pin_pwm 7 // LED1+PWM
#define Pin_dir 5 // DIRECCION
#define Pin_brk 6 // FRENO
#define Pin_led 13 // LED2

float ang = 0.0, angf = 0.0 ;
float uk = 0.0, uk1 = 0.0, yk1 = 0.0 ;
float ukr = 0.0, uk1r = 0.0, uk2r = 0.0, yk1r = 0.0, yk2r = 0.0 ;

// Definición de variables
unsigned long antMill = 0;  // ms anteriores
unsigned long antMillTX = 0;  // ms anteriores
float ref = 0.0; // Valor de referencia -> Posición Vertical de 90º
float u = 0.0; // Entrada/Retroalimentación del controlador
float err = 0.0; // Señal de error
float errant = 0.0; // Memoria de la señal de error anterior
float v = 0; // Salida de control
float integ = 0.0; // Almacenamiento para el integrador
float int_vel = 0.0; // Almacenamiento para el integrador para reducir velocidad

byte datos[2] = {0, 0};

void setup() {
  analogReadResolution(10); // Resolución del ADC -> 10 bits
  analogWriteResolution(12); // Resolución del DAC -> 12 bits

  pinMode(Pin_pwm, OUTPUT); // Salida -> LED1+PWM
  pinMode(Pin_dir, OUTPUT); // Salida -> DIRECCIÓN
  pinMode(Pin_led, OUTPUT); // Salida -> LED2
  pinMode(Pin_brk, OUTPUT); // Salida -> FRENO

  digitalWrite(Pin_brk, LOW);  // Activar freno

  Serial.begin(115200); // Inicializar puerto serie ordenador-Arduino
  SerialUSB.begin(256000); // Inicializar puerto serie ordenador-Simulink

  Serial.println("Inicio del control del péndulo"); // Mensaje inicial
}

void loop() {
  // Potenciometro 6127V1A180LS -> Salida de 0.2V a 4.8V - 0º a 180º
  lecturaPos();
  // Chequeo de señal de habilitado
  if ((datos[0] == 0)) {
    // Levantamiento del péndulo, a más de 35º se considera que ya ha caído
    if (abs(int(yk1)) > 35) {
      // Reinicio de las variables del proceso
      v = 0.0; integ = 0.0; int_vel = 0.0; err = 0.0;
      // Parada del motor antes de iniciar el levantamiento
      digitalWrite(Pin_brk, LOW); // Freno del motor
      delay(2500); // Espera 2.5s
      // Selección de dirección de giro dependiendo del lado al que se ha caído
      if (int(u) > 0) {
        digitalWrite(Pin_dir, LOW); // Dirección CW
        analogWrite(Pin_pwm, 0); // Motor a máxima velocidad
        digitalWrite(Pin_brk, HIGH); // Liberación del freno
        delay(3000); //Espera a que el motor esté a máxima velocidad
        digitalWrite(Pin_brk, LOW); // Freno del motor
        analogWrite(Pin_pwm, 2400); // Giro a velocidad reducida
        digitalWrite(Pin_dir, HIGH); // Dirección CCW
        delay(400); //Espera antes de liberar el freno
        digitalWrite(Pin_brk, HIGH); // Liberación del freno
        while (abs(int(yk1)) > 10) {
          lecturaPos();
          Serial.print("U_SR=");
          Serial.println(yk1);
        }
      }
      else {
        digitalWrite(Pin_dir, HIGH); // Dirección CCW
        analogWrite(Pin_pwm, 0); // Motor a máxima velocidad
        digitalWrite(Pin_brk, HIGH); // Liberación del freno
        delay(3000); //Espera a que el motor esté a máxima velocidad
        digitalWrite(Pin_brk, LOW); // Freno del motor
        analogWrite(Pin_pwm, 2300); // Giro a velocidad reducida
        digitalWrite(Pin_dir, LOW); // Dirección CW
        delay(400); //Espera antes de liberar el freno
        digitalWrite(Pin_brk, HIGH); // Liberación del freno
        while (abs(int(yk1)) > 10) {
          lecturaPos();
          Serial.print("U_SL=");
          Serial.println(yk1);
        }
      }
    }
    // Control PID seguimiento de referencia
    else
    {
      if (datos[1] != 0) {
        ref = 2 * float(sin(2 * PI * FREC * (millis() / 1000)));
      }
      else
        ref = 0.0;
      digitalWrite(Pin_brk, HIGH); // Liberación del freno
      // Vertical = 90º
      err = ref - yk1; // Error con referencia filtrada
      integ += err * Tsc; // Cálculo de la acción integral

      // Saturación de la acción integral
      if (int(integ) > 2400)
        integ = VMAX;
      else if (int(integ) < -2400)
        integ = -VMAX;

      // Reducción de velocidad en posición vertical
      if ((int(datos[1]) == 0) && ((abs(int(err * 100)) < 70) && (abs(int(v)) > 50)) ) {
        int_vel += v * Kc;
        // Saturación del integrador
        if (int(int_vel) > 3000)
          int_vel = 3000;
        else if (int(int_vel) < -3000)
          int_vel = -3000;
      }
      else {
        // Reducción del integrador para evitar influencias en el PID
        if (abs(int(int_vel)) > 10)
          int_vel = int_vel * 0.999;
      }

      v = err * Kp + integ * Ki - int_vel; // Salida
      errant = err; // Guardar el error para la próxima iteración

      if (int(v) > 2400)
        v = VMAX;
      else if (int(v) < -2400)
        v = -VMAX;
      // Escritura de acción de control y dirección de giro
      analogWrite(Pin_pwm, 4095 - map(abs(int(v)), 0, 2400, 0, 4095));
      if (int(v) >= 0)
        digitalWrite(Pin_dir, LOW); // Dirección CW
      else
        digitalWrite(Pin_dir, HIGH); // Dirección CCW
    }
  }
  else {
    digitalWrite(Pin_brk, LOW); // Freno del motor
    analogWrite(Pin_pwm, 4095); // Velocidad de motor cero
  }
  // Envío de datos al ordenador - Arduino (Monitor serie)
  serialMon();

  // Envío de datos al ordenador - Arduino (Solo serial plotter)
  //serialPlot();

  // Envío/recepción de datos al ordenador - Simulink cada 100ms
  unsigned long actMillTX = millis();
  if (actMillTX - antMillTX > 100) {
    simulinkTXRX();
    antMillTX = actMillTX;
  }
  // Monitorea la duración del control. Si se pasa del Ts enciende LED2
  unsigned long actMill = millis();  // Obtiene los ms actuales

  if (actMill - antMill > Ts) {
    digitalWrite(Pin_led, HIGH);
    Serial.println("-----Fuera de tiempo!"); // Transmite mensaje de fallo
  }
  else
    digitalWrite(Pin_led, LOW);

  antMill = actMill; // Almacena los ms actuales para la próxima iteración
}

void lecturaPos() {
  u = float(map(analogRead(Pin_pend), 60, 1024, -7000, 7000)) / 100.0; // Lectura de posición
  uk1 = uk ; uk = u ;
  angf = 0.2929 * uk + 0.2929 * uk1 + 0.4142 * yk1 ; //wn=0.25 ;
  yk1 = angf ;
}

void serialMon() {
  Serial.print(millis());
  Serial.print(';');
  Serial.print(yk1);
  Serial.print(';');
  Serial.print(ref);
  Serial.print(';');
  Serial.println(v);
}

void serialPlot() {
  Serial.print(ref * 100);
  Serial.print(',');
  Serial.print(yk1 * 100);
  Serial.print(',');
  Serial.print(int_vel);
  Serial.print(',');
  Serial.println(v);
}

void simulinkTXRX() {
  // Envío y recepción de datos de Simulink
  // En este caso sólo se deben reciben 0 y 1
  if (SerialUSB.available() >= 1) {
    int bb = 0;
    while (bb != 255 && bb != -1)
      bb = SerialUSB.read();

    if (SerialUSB.available() >= 2) {
      datos[0] = map(SerialUSB.read(), 0, 254, DATMIN * 100, DATMAX * 100) / 100.0; // Señal de referencia
      datos[1] = map(SerialUSB.read(), 0, 254, DATMIN * 100, DATMAX * 100) / 100.0; // Señal de parada del proceso 0-Funcionando / 1-Detenido
    }
    while (bb != -1)
      bb = SerialUSB.read();

    SerialUSB.write(255);
    SerialUSB.write(map(int(yk1 * 10), -700, 700, 0, 254));
    SerialUSB.write(map(int(v), -2400, 2400, 0, 254));
  }
}
